# -*- coding: utf-8 -*-
"""
This package collects utility functions used throughout the ``ChemPy`` package.
"""
from __future__ import (absolute_import, division, print_function)

from .pyutil import NoConvergence
