"""
This package collects functions useful for studying chemical kinetics problems.
"""
from __future__ import (absolute_import, division, print_function)

from .rates import MassAction, EyringHS
from .eyring import EyringParam
from .arrhenius import ArrheniusParam
